#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "niflib.h"
#include "obj/NiObject.h"
#include "obj/NiAVObject.h"
#include "obj/NiNode.h"
#include "obj/NiProperty.h"
#include "obj/NiMaterialProperty.h"
#include "obj/NiTexturingProperty.h"
#include "obj/NiSourceTexture.h"
#include "obj/NiTriBasedGeom.h"
#include "obj/NiTriBasedGeomData.h"
#include "obj/NiBinaryExtraData.h"
#include "obj/NiTriStripsData.h"

#include <strstream>
#include <vector>

using namespace Niflib;

#define SAFERELEASEP(a) { if(a) { delete[] a; } a=0; }

struct DXVertex {
	float Position[3];
	float Normal[3];
	float TexCoord[2];
};
struct DXMatrix {
	float data[4*4];
};

struct ExportedNode {
	int verts;
	int faces;
	Matrix44 transform;
	DXVertex* vBuffer;
	unsigned short* iBuffer;
	char* tex;

	ExportedNode() { ZeroMemory(this,sizeof(ExportedNode)); }

	~ExportedNode() {
		SAFERELEASEP(vBuffer);
		SAFERELEASEP(iBuffer);
		SAFERELEASEP(tex);
	}
};

static NiAVObjectRef LoadStreamNif(const char* data, int size) {
	istrstream s(data, size);
	NiAVObjectRef rootObj = DynamicCast<NiAVObject>(ReadNifTree( s, 0 ));
	return rootObj;
}
static void SearchShapes( NiAVObjectRef rootObj, vector<NiTriBasedGeomRef>* SubsetNodes ) {
    //Check if this object is derived from NiTriBasedGeom
    NiTriBasedGeomRef niGeom = DynamicCast<NiTriBasedGeom>( rootObj );
    if(niGeom) {
		SubsetNodes->push_back(niGeom);
	} else {
		//Check if this object derives from NiNode and, thus, may have children
		NiNodeRef niNode = DynamicCast<NiNode>( rootObj );
		if(niNode) {
			//Call this function for any children
			vector<NiAVObjectRef> children = niNode->GetChildren();
			for(uint i = 0; i < children.size(); i++) {
				SearchShapes( children[i], SubsetNodes );
			}
		}
	}
}

static bool ExportShape( NiTriBasedGeomRef niGeom, ExportedNode* node, float* maxsize ) {
	//Check that an external texture exists
	NiTexturingPropertyRef niTexProp = DynamicCast<NiTexturingProperty>(niGeom->GetPropertyByType( NiTexturingProperty::TypeConst() ));
	if(!niTexProp||niTexProp->GetTextureCount()==0) return false;
	TexDesc texDesc=niTexProp->GetTexture(0);
	NiSourceTextureRef niSrcTex = texDesc.source;
	if(!niSrcTex || !niSrcTex->IsTextureExternal() ) return false;

	//Get data object (NiTriBasedGeomData) from geometry node (NiTriBasedGeom)
	NiTriBasedGeomDataRef niGeomData = DynamicCast<NiTriBasedGeomData>(niGeom->GetData());
	if(!niGeomData) return false;

	//Check that there is at least one set of texture coords available
	if(niGeomData->GetUVSetCount()==0) return false;

	//Indicies
	vector<Triangle> tris = niGeomData->GetTriangles();
	node->faces=tris.size();
	if(node->faces==0) return false;

	//Calculate mesh bounding sphere
	float radius = niGeomData->GetRadius();
	Vector3 center = niGeomData->GetCenter();
	radius = radius + sqrt(center.x*center.x + center.y*center.y + center.z*center.z);
	if(radius>*maxsize) *maxsize=radius;

	//Now that we're sure this mesh is valid, start the conversion
	node->iBuffer=new unsigned short[node->faces*3];
	for(int i=0;i<node->faces;i++) {
		node->iBuffer[i*3+0]=tris[i].v1;
		node->iBuffer[i*3+1]=tris[i].v2;
		node->iBuffer[i*3+2]=tris[i].v3;
	}

	//Vertex data
	vector<Vector3> positions = niGeomData->GetVertices();
	vector<Vector3> normals = niGeomData->GetNormals();
	vector<TexCoord> texCoords = niGeomData->GetUVSet(0);

	//Verticies
	node->verts=niGeomData->GetVertexCount();

	node->vBuffer=new DXVertex[node->verts];
	for(int i=0;i<node->verts;i++) {
		node->vBuffer[i].Position[0]=positions[i].x;
		node->vBuffer[i].Position[1]=positions[i].y;
		node->vBuffer[i].Position[2]=positions[i].z;
		if(normals.size()>0) {
			node->vBuffer[i].Normal[0]=normals[i].x;
			node->vBuffer[i].Normal[1]=normals[i].y;
			node->vBuffer[i].Normal[2]=normals[i].z;
		} else {
			node->vBuffer[i].Normal[0]=0;
			node->vBuffer[i].Normal[1]=0;
			node->vBuffer[i].Normal[2]=1;
		}
		node->vBuffer[i].TexCoord[0]=texCoords[i].u;
		node->vBuffer[i].TexCoord[1]=texCoords[i].v;
	}

	//Get texture file path
	string s=niSrcTex->GetExternalFileName();
	node->tex=new char[s.length()+1];
	strcpy(node->tex,s.c_str());

	//Get transformation of mesh as 4x4 radius
	node->transform = niGeom->GetWorldTransform();

	return true;
}

static bool ResaveNif(const char* outpath, ExportedNode* nodes, int nodeCount, float maxsize, byte Far) {
	HANDLE h=CreateFileA(outpath, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0);
	DWORD unused;
	if(h==INVALID_HANDLE_VALUE) return false;
	
	WriteFile(h, &nodeCount, 4, &unused, 0);
	WriteFile(h, &maxsize, 4, &unused, 0);
	WriteFile(h, &Far, 1, &unused, 0);
	for(int i=0;i<nodeCount;i++) {
		_strlwr(nodes[i].tex);
		short slen=strlen(nodes[i].tex) + 1;

		WriteFile(h, &nodes[i].transform, sizeof(Matrix44), &unused, 0);
		WriteFile(h, &nodes[i].verts, 4, &unused, 0);
		WriteFile(h, &nodes[i].faces, 4, &unused, 0);
		WriteFile(h, nodes[i].vBuffer, nodes[i].verts*sizeof(DXVertex), &unused, 0);
		WriteFile(h, nodes[i].iBuffer, nodes[i].faces*6, &unused, 0);
		WriteFile(h, &slen, 2, &unused, 0);
		WriteFile(h, nodes[i].tex, slen, &unused, 0);
	}
	CloseHandle(h);

	return true;
}
float _stdcall ProcessNif(char* data, int datasize, char* outpath, float cutoff, float cutoff2) {
	vector<NiTriBasedGeomRef> SubsetNodes;
	int nodesExported=0;
	int numNodes;
	ExportedNode* nodes;
	float maxsize=0;

	NiAVObjectRef rootObj=LoadStreamNif(data, datasize);
	if(!rootObj) return -1;
	
	SearchShapes(rootObj, &SubsetNodes);
	numNodes=SubsetNodes.size();
	if(!numNodes) return -1;
	
	nodes = new ExportedNode[numNodes];
	for(int i = 0;i<numNodes;i++) {
		if(ExportShape(SubsetNodes[i], &nodes[i], &maxsize)) {
			nodesExported++; 
		} else {
			i--;
			numNodes--;
		}
	}

	SubsetNodes.clear();

	if(!numNodes||maxsize<cutoff) {
		delete[] nodes;
		return -1;
	}

	if(!outpath) {
		delete[] nodes;
		return maxsize;
	}

	bool result=ResaveNif(outpath, nodes, numNodes, maxsize, (maxsize>cutoff2)?1:0);

	delete[] nodes;

	if(!result) return -1; else return maxsize;
}