using System;
using FileVersionInfo=System.Diagnostics.FileVersionInfo;
using Process=System.Diagnostics.Process;
using File=System.IO.File;
using Directory=System.IO.Directory;
using Serializer=System.Runtime.Serialization.Formatters.Binary.BinaryFormatter;
using Application=System.Windows.Forms.Application;
using MessageBox=System.Windows.Forms.MessageBox;
using MessageBoxButtons=System.Windows.Forms.MessageBoxButtons;
using DialogResult=System.Windows.Forms.DialogResult;
using MGEgui.DirectX;

namespace MGEgui {

    [Serializable]
    public class Technique {
        public string Description;
        public string Name;
        public string[] Lines;

        public Technique() {
            Description="Description";
            Name="Unnamed";
            Lines=new string[0];
        }
    }

    /// <summary>
    /// The list of different macro types
    /// </summary>
    /// <remarks>
    /// Copied from input.h
    /// Make sure that this is always up to date (Needs Console3, Console4, Offhammer)
    /// </remarks>
    public enum MacroType : byte {
        Unused=0,       //This key isn't assigned to anything
        //FD_Console
        Console1=1,     //Enter console command, exit console
        Console2=2,     //Dont open or close console
        //FD_Press
        Hammer1=3,      //Hammer a key while pressed
        Hammer2=4,      //Start hammering a key
        Unhammer=5,     //Stop hammering a key
        AHammer1=6,     //As Hammer except hammers in alternate frames
        AHammer2=7,
        AUnhammer=8,
        Press1=9,       //Other keys pressed while this is pressed
        Press2=10,       //Press keys when this is pressed, but dont depress
        Unpress=11,      //Depress any pressed keys
        //FD_Timer
        BeginTimer=12,   //Start a timer trigger
        EndTimer=13,    //End a timer trigger
        //FD_Graphics
        Graphics=14     //Perform a graphics function
    }

    /// <summary>
    /// What a particular MacroEditerForm is being used for
    /// </summary>
    public enum MacroFormType { Unknown, Editor, Console, Press, Function, Timer, Trigger }

    [Serializable]
    public struct KeyPress {
        public bool Down;
        public byte Code;

        public KeyPress(byte code,bool down) {
            Down=down;
            Code=code;
        }
    }

    [Serializable]
    public class FD_Console {
        public byte Length;             //The size of the string
        public KeyPress[] KeyCodes;     //A list of keycodes
        public string Description;      //A description of the command
        public FD_Console() {
            Length=0;
            KeyCodes=new KeyPress[Statics.MACROS];
            Description="";
        }
    }
    [Serializable]
    public class FD_Press {
        public bool[] KeyStates;    //Last 8 for mouse
        public FD_Press() {
            KeyStates=new bool[Statics.MACROS];
        }
    }
    [Serializable]
    public class FD_Timer {
        public byte TimerID;           //The timer to activate/deactivate
        public FD_Timer() {
            TimerID=0;
        }
    }
    [Serializable]
    public class FD_Graphics {
        public byte function; //The GraphicsFuncs to call in the fake d3d dll
    }

    [Serializable]
    public class Macro {
        public MacroType Type;
        public FD_Console Console;
        public FD_Press Press;
        public FD_Timer Timer;
        public FD_Graphics Graphics;

        public Macro() {
            Console=new FD_Console();
            Press=new FD_Press();
            Timer=new FD_Timer();
            Graphics=new FD_Graphics();
        }
    }
    [Serializable]
    public class Trigger {
        public uint TimeInterval;
        public bool Active;
        public FD_Press data;

        public Trigger() {
            TimeInterval=0;
            Active=false;
            data=new FD_Press();
        }
    }

    public static class MutexCheck {
        private const string MutexName="MGEguiMutex";
        private static System.Threading.Mutex mutex;

        internal static bool PerformCheck() {
            bool n;
            mutex=new System.Threading.Mutex(true, MutexName, out n);
            return n;
        }
    }

    /// <summary>
    /// Contains some constant variables, misc static functions and the program entry point
    /// </summary>
    public class Statics {
        /// <summary>
        /// Used to ensure that created options files match the dll versions
        /// </summary>
        public const byte MajorVersion=3;
        public const byte MinorVersion=3;
        public const uint BuildVersion=1;
        public const string VersionString="3.3.1";
        public const string versionString="v"+VersionString;
        public const byte SaveVersion=43;
        public const byte DistantLandVersion=0;

        public const int MACROS=266;
        public const int TRIGGERS=4;

        public static Macro[] Macros=new Macro[MACROS];
        public static Trigger[] Triggers=new Trigger[TRIGGERS];
        public static byte[] Remapper=new byte[256];

        public static Serializer formatter=new Serializer();

        private struct CommandArgs {
            public readonly bool styles;
            public readonly bool mutex;

            public CommandArgs(string[] args) {
                styles=true;
                mutex=true;
                foreach(string s in args) {
                    switch(s) {
                        case "-nostyles":
                            styles=false;
                            break;
                        case "-nomutex":
                            mutex=false;
                            break;
                    }
                }
            }
        }

        /// <summary>
        /// Entry point for this program
        /// </summary>
        [STAThread]
        public static void Main(string[] sargs) {
            CommandArgs args=new CommandArgs(sargs);

            if(args.styles) Application.EnableVisualStyles();

            if(args.mutex&&!MutexCheck.PerformCheck()) {
                MessageBox.Show("MGEgui.exe is already running.", "Error");
                return;
            }

            Directory.SetCurrentDirectory(System.IO.Path.GetDirectoryName(Application.ExecutablePath));
            if(!File.Exists("./morrowind.exe")||!File.Exists("./morrowind.ini")||!Directory.Exists("data files")) {
                MessageBox.Show("MGE must be installed to the morrowind directory.","Error");
                return;
            }
			if(!Directory.Exists("MGE3")||!File.Exists("./MGE3/MGEfuncs.dll")||!File.Exists("./d3d8.dll")||
               !File.Exists("./dinput8.dll")) {
                MessageBox.Show("One or more of MGE's files appear to be missing. Please reinstall","Error");
                return;
            }
            //Morrowind version info
            try {
                FileVersionInfo MorrowVersion=FileVersionInfo.GetVersionInfo("Morrowind.exe");
                if(MorrowVersion.FileMajorPart!=1||MorrowVersion.FileMinorPart<6) {
                    MessageBox.Show("Your version of morrowind is not compatable with MGE.\n"+
                        "MGE requires a fully patched copy of bloodmoon. (i.e. morrowind version 1.6.1820)", "Error");
                    return;
                }
            } catch {
                MessageBox.Show("Morrowind.exe appears to be corrupt.\n"+
                    "MGE is unable to determine Morrowinds version.", "Error");
                return;
            }
            //Check for dsound.dll
            if(File.Exists("dsound.dll")&&new System.IO.FileInfo("dsound.dll").Length==16384) {
                try {
                    File.Delete("dsound.dll");
                } catch {
                    MessageBox.Show("You appear to have part of an old version of MGE installed.\n"+
                        "You need to remove dsound.dll from your morrowind directory for this version of MGE to function.", "Error");
                    return;
                }
            }
            //Create the data files directories MGE uses
            if(!Directory.Exists(@"data files\shaders")) {
                Directory.CreateDirectory(@"data files\shaders");
            }
            if(!Directory.Exists(@"data files\MGE meshes")) {
                Directory.CreateDirectory(@"data files\MGE meshes");
            }
            if(!Directory.Exists(@"data files\Textures\MGE")) {
                Directory.CreateDirectory(@"data files\Textures\MGE");
            }
            if(!Directory.Exists(@"data files\MGE videos")) {
                Directory.CreateDirectory(@"data files\MGE videos");
            }
            if(!Directory.Exists(@"data files\shaders\default")) {
                Directory.CreateDirectory(@"data files\shaders\default");
            }
            if(!Directory.Exists(@"data files\shaders\water")) {
                Directory.CreateDirectory(@"data files\shaders\water");
            }
            //Create some structures
            for(int i=0;i<MACROS;i++) {
                Macros[i]=new Macro();
            }
            for(int i=0;i<TRIGGERS;i++) {
                Triggers[i]=new Trigger();
            }
            DXMain.GetDeviceCaps();
            MainForm mf=new MainForm();
            Application.Run(mf);
        }

    }
}
